-- Title       : Counter_N_RHSEU
-- Design      : Interfaces
-- Author      : NC

-------------------------------------------------------------------------------
--
-- Description :  Triple Redundancy on States 
--				  Protection against SEU in Flip-Flop
--				  Counter_N is a Moore FSM
-------------------------------------------------------------------------------


LIBRARY IEEE;
USE IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all; 
use IEEE.std_logic_arith.all;
use work.interfaces_pkg.all;

entity  Counter_N_RHSEU is
	generic(
      N : integer := 8; -- number of bits
      modul : integer := 256  -- modulus of the counter
	  );                                -- if modul is set to 0, the counter free-runs up to 2^N-1 
	port(clk   : in  std_logic;	-- input signal to be counted  
		rst_L  : in  std_logic; -- active low reset	  
		en  : in  std_logic; -- active high enable
		count   : out  std_logic_VECTOR (N-1 downto 0); -- output count value 
		seu	: out std_logic);
end Counter_N_RHSEU;													  

architecture Counter_N1 of Counter_N_RHSEU is
signal next_count, count_a, count_b, count_c, count_corr: std_logic_VECTOR (N-1 downto 0);
signal a_eq_b: std_logic;
signal b_eq_c : std_logic; 
signal a_eq_c : std_logic;
signal status :  STD_LOGIC_VECTOR(2 downto 0);	
	
	

begin

 gen_mod : if not(modul = 0) generate

	 fsm_state: process (clk)
		begin
			if rising_edge(clk) then 
				if rst_L = '0' then	
					count_a <= (others => '0');
					count_b <= (others => '0');
					count_c <= (others => '0');	
				else	 
					count_a <= next_count;
					count_b <= next_count;
					count_c <= next_count;
				end if;
		end if;
	end process; 
	
	process (count_a, count_b, count_c)
 	begin
		
		if (count_a = count_b) then
			a_eq_b <= '1';
		else
			a_eq_b <= '0';
		end if;	
			
		if (count_a = count_c) then
			a_eq_c <= '1';
		else
			a_eq_c <= '0';
		end if;		
					
		if (count_b = count_c) then
			b_eq_c <= '1';
		else
			b_eq_c <= '0';
		end if;
  end process;
  
  status <= a_eq_b & a_eq_c & b_eq_c;
  
  process (status, count_a, count_b)
    begin
		case (status) is 
			when "111" =>     			--All equal
				count_corr <= count_a; 
				seu <= '0'; 
			when "100" =>    			--c different
				count_corr <= count_a; 
				seu <= '1'; 
			when "010" => 				--b different
				count_corr <= count_a;
				seu <= '1'; 
			when "001" => 				--a different
				count_corr <= count_b;
				seu <= '1'; 
			when others => 			  -- errors in majority logic
				count_corr <= count_a;
				seu <= '0';
		end case;
	end process;  
	
	Count_incr_logic: process(count_corr, en)		
	begin	
   	 if en = '1' then
			if (count_corr = CONV_STD_LOGIC_VECTOR((modul-1), N)) then
    			next_count <= (others => '0');
			else 
				next_count <= count_corr + 1;
			end if;
      else
        next_count <= count_corr;
	  end if;
	end process;  
	
    count <= count_corr;
    
   end generate gen_mod;
  

  gen_no_mod : if (modul = 0) generate
  
    fsm_state: process (clk)
		begin
			if rising_edge(clk) then 
				if rst_L = '0' then	
					count_a <= (others => '0');
					count_b <= (others => '0');
					count_c <= (others => '0');	
				else	 
					count_a <= next_count;
					count_b <= next_count;
					count_c <= next_count;
				end if;
		end if;
	end process; 
	
	process (count_a, count_b, count_c)
 	begin
		
		if (count_a = count_b) then
			a_eq_b <= '1';
		else
			a_eq_b <= '0';
		end if;	
			
		if (count_a = count_c) then
			a_eq_c <= '1';
		else
			a_eq_c <= '0';
		end if;		
					
		if (count_b = count_c) then
			b_eq_c <= '1';
		else
			b_eq_c <= '0';
		end if;
  end process;
  
  status <= a_eq_b & a_eq_c & b_eq_c;
  
  process (status, count_a, count_b)
    begin
		case (status) is 
			when "111" =>     			--All equal
				count_corr <= count_a; 
				seu <= '0'; 
			when "100" =>    			--c different
				count_corr <= count_a; 
				seu <= '1'; 
			when "010" => 				--b different
				count_corr <= count_a;
				seu <= '1'; 
			when "001" => 				--a different
				count_corr <= count_b;
				seu <= '1'; 
			when others => 			  -- errors in majority logic	
				count_corr <= count_a;
				seu <= '1';
				null; 
		end case;
	end process;  
	
	Count_incr_logic: process(count_corr, en)		
	begin	
      if en = '1' then
          next_count <= count_corr + 1;
      else
          next_count <= count_corr;
      end if;
	end process;
    
    count <= count_corr;
    
  end generate gen_no_mod;

  
END Counter_N1;