-------------------------------------------------------------------------------
--
-- Title       : FIFO_N_D
-- Design      : Interfaces
-- Author      : Cla
-- Company     : Casa
--
-- Description :  N-bit, D-locations FIFO buffer
--
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
USE IEEE.std_logic_arith.all;
use work.interfaces_pkg.all;

entity FIFO_N_D is
	generic (N, D :	integer);
	port(
		word_in : in STD_LOGIC_VECTOR(N-1 downto 0);
		write : in STD_LOGIC;
		clk : in std_logic;
		read : in STD_LOGIC;
		rst_L : in std_logic;
		word_out : out std_logic_vector(N-1 downto 0);
		full : out STD_LOGIC;
		empty : out STD_LOGIC
		);
end FIFO_N_D;

architecture FIFO_N_D1 of FIFO_N_D is
	
	type fifo_array_type is array (D-1 downto 0) of std_logic_vector(N-1 downto 0);
	
	signal fifo_array : fifo_array_type;
	signal W_pnt : INTEGER range 0 to D-1; -- write pointer
	signal R_pnt : INTEGER range 0 to D-1;  -- read pointer
	signal empty_int : std_logic;
	signal full_int : std_logic;
	
	
	
begin
	
	process (clk) -- write process
	begin
		
		if rising_edge(clk) then
			if rst_L = '0' then
				for INDEX in 0 to D-1 loop
					fifo_array(INDEX) <= (others => '0');
				end loop;
			elsif write = '1' and full_int = '0' then
				fifo_array(W_pnt) <= word_in;
			end if;
		end if;
		
	end process;
	
	process (CLK)  -- pointer updating process
		variable w_p, r_p : INTEGER range 0 to D-1; -- write/read pointer local variables  
		variable ocpy : INTEGER range 0 to D; -- occupancy
	begin
		
		if rising_edge(CLK) then
			if rst_L = '0' then
				W_pnt <= 0;
				R_pnt <= 0;
				empty_int <= '1';
				full_int <= '0';
				ocpy := 0;
			else 
				if write = '1' and full_int = '0' then
					if W_pnt < D-1 then
						w_p := W_pnt + 1;
					elsif W_pnt = D-1 then
						w_p := 0;
					end if;
					ocpy := ocpy + 1;
					W_pnt <= w_p;
					if (read /= '1') OR (empty_int = '1') then -- empty and full flags must be updated only if a reading has been actually performed
						empty_int <= '0';
						if w_p = R_pnt then
							full_int <= '1';
						end if;
					end if;
				end if;
				if read = '1' and empty_int = '0' then 
					if R_pnt < D-1 then
						r_p := R_pnt + 1;
					elsif R_pnt = D-1 then
						r_p := 0;
					end if;
					ocpy := ocpy - 1;
					R_pnt <= r_p;
					if (write /= '1') OR (full_int = '1') then -- empty and full flags must be updated only if a writing has been actually performed
						full_int <= '0';
						if r_p = W_pnt then
							empty_int <= '1';
						end if;
					end if;
				end if;
			end if;
		end if;
		
	end process;
	
	
	word_out <= fifo_array(R_pnt);
	full <= full_int;
	empty <= empty_int;
	
end FIFO_N_D1;
