-------------------------------------------------------------------------------
--
-- Title       : FIFO_NxD_RHSEU
-- Design      : Interfaces
-- Author      : Cla
-- Company     : Casa
--
-- Description :  N-bit, D-locations FIFO buffer
--
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
USE IEEE.std_logic_arith.all;
USE IEEE.std_logic_misc.all;
USE IEEE.std_logic_unsigned.all;
USE IEEE.math_real.all;
use work.interfaces_pkg.all;

entity FIFO_NxD_RHSEU is
	generic (
	  N : integer := 8; 
	  D : integer := 16
	);
	port(
		word_in : in STD_LOGIC_VECTOR(N-1 downto 0);
		write : in STD_LOGIC;
		clk : in std_logic;
		read : in STD_LOGIC;
		rst_L : in std_logic;
		word_out : out std_logic_vector(N-1 downto 0);
		full : out STD_LOGIC;
		empty : out STD_LOGIC;
		seu : out STD_LOGIC
		);
end FIFO_NxD_RHSEU;

architecture FIFO_NxD_RHSEU1 of FIFO_NxD_RHSEU is
	
  component reg_N_RHSEU is
	generic (N : integer );
	port (
		data_in : in std_logic_vector(N-1 downto 0);
		clk : in std_logic;
		rst_L : in std_logic; -- active low reset
		write : in std_logic; -- write command
		data_out : out std_logic_vector(N-1 downto 0);
		seu: out std_logic
	);
  end component;  
  
    
  component ffd_RHSEU is
      port (
          CLR : in std_logic; -- active low reset
          en : in std_logic; -- en command
          clk : in std_logic;
          D : in std_logic;
          Q : out std_logic;
          QN : out std_logic;
          seu: out std_logic
      );
  end component;
    
    constant D_real :real := real(D);
    constant pnt_width_real : real := log(real(D))/log(real(2)); 
    constant pnt_width : integer := integer(pnt_width_real); 
      
	type fifo_array_type is array (D-1 downto 0) of std_logic_vector(N-1 downto 0);
	
	signal w_pnt : std_logic_vector(pnt_width-1 downto 0); -- write pointer
	signal r_pnt : std_logic_vector(pnt_width-1 downto 0);  -- read pointer
	signal w_pnt_in : std_logic_vector(pnt_width-1 downto 0); 
	signal r_pnt_in : std_logic_vector(pnt_width-1 downto 0);
	signal w_pnt_wren : std_logic;
	signal r_pnt_wren : std_logic; 
	signal empty_out_int : std_logic;
	signal full_out_int : std_logic; 
	signal empty_in : std_logic;
	signal full_in : std_logic; 
	signal empty_wren : std_logic;
	signal full_wren : std_logic;
	
	signal regs_wr : std_logic_vector(D-1 downto 0);
	signal seu_REGS : std_logic_vector(D-1 downto 0);
	signal seu_W_PNT_REG : std_logic;
	signal seu_R_PNT_REG : std_logic;
	signal seu_FULL_FF : std_logic;
	signal seu_EMPTY_FF : std_logic;
	signal regs_out : fifo_array_type;
	
	
begin

  regs_gen: for i in 0 to D-1 generate 
		
    REG : reg_N_RHSEU
    generic map (
      N => N
    )  
    port map (
       data_in  => word_in
      ,clk      => clk
      ,rst_L    => rst_L
      ,write    => regs_wr(i)
      ,data_out => regs_out(i)
      ,seu      => seu_REGS(i)
    );
		
  end generate regs_gen;

  
  W_PNT_REG : reg_N_RHSEU
  generic map (
    N => pnt_width
  )  
  port map (
     data_in  => w_pnt_in
    ,clk      => clk
    ,rst_L    => rst_L
    ,write    => w_pnt_wren
    ,data_out => w_pnt
    ,seu      => seu_W_PNT_REG
  );
    
    
  R_PNT_REG : reg_N_RHSEU
  generic map (
    N => pnt_width
  )  
  port map (
     data_in  => r_pnt_in
    ,clk      => clk
    ,rst_L    => rst_L
    ,write    => r_pnt_wren
    ,data_out => r_pnt
    ,seu      => seu_R_PNT_REG
  );
  
  FULL_FFD : ffd_rhseu
	port map(
		CLR => rst_L,
		en => full_wren,
		clk => clk,
		D => full_in,
		Q => full_out_int,
		QN => open,
		seu => seu_FULL_FF
	);
  
  EMPTY_FFD : ffd_rhseu
	port map(
		CLR => rst_L,
		en => empty_wren,
		clk => clk,
		D => empty_in,
		Q => empty_out_int,
		QN => open,
		seu => seu_EMPTY_FF
	);
  
  process (write, full_out_int, w_pnt) -- write process
  begin
    if write = '1' and full_out_int = '0' then
      for i in 0 to D-1 loop
        if (CONV_INTEGER(unsigned(w_pnt)) = i) then
          regs_wr(i) <= '1';
        else 
          regs_wr(i) <= '0';
        end if;
      end loop;
    else
      regs_wr <= (OTHERS => '0');
    end if;
  end process;
	
    
    
  process (write, read, full_out_int, empty_out_int, w_pnt, r_pnt)  -- pointer updating process
      variable w_p, r_p : std_logic_vector(pnt_width-1 downto 0); -- write/read pointer local variables
      
  begin
      
    w_pnt_in   <= (OTHERS => '0');
    w_pnt_wren <= '0'; 
    r_pnt_in   <= (OTHERS => '0');
    r_pnt_wren <= '0';      
    full_in    <= '0';
    full_wren  <= '0';       
    empty_in   <= '0';
    empty_wren <= '0'; 
    if (write = '1') and (full_out_int = '0') then
      if CONV_INTEGER(unsigned(w_pnt)) < D-1 then
          w_p := w_pnt + 1;
      else
          w_p := (OTHERS => '0');
      end if;
      w_pnt_in <= w_p;
      w_pnt_wren <= '1';
      if (read = '0') OR (empty_out_int = '1') then -- a read command can prevent the empty and full flags to be updated only if a reading has been actually performed
        empty_in <= '0';
        empty_wren <= '1';
        if w_p = r_pnt then
            full_in <= '1';
            full_wren <= '1';
        end if;
      end if;
    end if;
    if (read = '1') and (empty_out_int = '0') then 
      if CONV_INTEGER(unsigned(r_pnt)) < D-1 then
          r_p := r_pnt + 1;
      else
          r_p := (OTHERS => '0');
      end if;
      r_pnt_in <= r_p;
      r_pnt_wren <= '1';
      if (write = '0') OR (full_out_int = '1') then -- a write command can prevent the empty and full flags to be updated only if a writing has been actually performed
          full_in <= '0';
          full_wren <= '1';
          if r_p = w_pnt then
            empty_in <= '1';
            empty_wren <= '1';
          end if;
      end if;
    end if;
      
  end process;
  
  
  word_out <= regs_out(CONV_INTEGER(unsigned(r_pnt)));
  full <= full_out_int;
  empty <= empty_out_int;
  
  seu <= seu_W_PNT_REG or seu_R_PNT_REG or seu_FULL_FF or seu_EMPTY_FF or or_reduce(seu_REGS);
	
end FIFO_NxD_RHSEU1;
