-- Title       : HDR_manager_RHSEU
-- Design      : Interfaces
-- Author      : NC

-------------------------------------------------------------------------------
--
-- Description :  Triple Redundancy on States and Output Registers
--				  Protection against SEU in Flip-Flop
--				  HDR_manager is a registered Mealy FSM	(all registered outputs)
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity HDR_manager_RHSEU is 
	port (
		clk: in STD_LOGIC;
		HDR_block: in STD_LOGIC;
		HDR_in: in STD_LOGIC;
		rst_L: in STD_LOGIC;
		HDR_wait: out STD_LOGIC;
		seu: out STD_LOGIC
	);
end HDR_manager_RHSEU;

architecture HDR_manager1 of HDR_manager_RHSEU is	 
type hdr_man_states is (IDLE, WAITING);	
signal next_state, current_state_a, current_state_b, current_state_c, current_state_corr: hdr_man_states;
signal status :  STD_LOGIC_VECTOR(2 downto 0);
signal HDR_wait_int: STD_LOGIC;
signal HDR_wait_sig:  STD_LOGIC_VECTOR(0 to 2);
signal corr_data_HDR : STD_LOGIC;
signal a_eq_b: std_logic;
signal b_eq_c : std_logic; 
signal a_eq_c : std_logic;	
signal seu_states : std_logic;	
signal seu_HDR_wait : std_logic;	



begin

fsm_state: process (clk)
begin
	if rising_edge(clk) then 
		if rst_L = '0' then	
			current_state_a <= IDLE;
			current_state_b <= IDLE;
			current_state_c <= IDLE;
		
		for i in 0 to 2 loop
			HDR_wait_sig(i) <= '0';
		end loop; 
		
		else	 
			current_state_a <= next_state;
			current_state_b <= next_state;
			current_state_c <= next_state;
			
		for i in 0 to 2 loop
			HDR_wait_sig(i) <= HDR_wait_int;
		end loop;
			
		end if;
	end if;
end process;

fsm_comb_logic: process(current_state_corr, HDR_block, HDR_in)
			
	   begin
		case current_state_corr is
				when IDLE =>
					if (HDR_in = '1') AND (HDR_block = '1') then	
						next_state <= WAITING;
						HDR_wait_int <= '1';						
					else
						next_state <= IDLE;
						HDR_wait_int <= '0';
					end if;
  
				when WAITING =>
						if HDR_block = '1' then	
						next_state <= WAITING;
						HDR_wait_int <= '1';
					else
						next_state <= IDLE;
						HDR_wait_int <= '0';
					end if;
				when others =>
						next_state <= IDLE;
						HDR_wait_int <= '0';
			end case;
		end process;  
		
	
	
  process (current_state_a, current_state_b, current_state_c)
	begin
        
		if (current_state_a = current_state_b) then
			a_eq_b <= '1';
		else
			a_eq_b <= '0';
		end if;
			
		if (current_state_a = current_state_c) then
			a_eq_c <= '1';
		else
			a_eq_c <= '0';
		end if;
				
		if (current_state_b = current_state_c) then
			b_eq_c <= '1';
		else
			b_eq_c <= '0';
		end if;
  end process;
  
  status <= a_eq_b & a_eq_c & b_eq_c;
  
  process (status, current_state_a, current_state_b)
    begin
		case (status) is 
			when "111" =>     			--All equal
				current_state_corr <= current_state_a; 
				seu_states <= '0'; 
			when "100" =>    			--c different
				current_state_corr <= current_state_a; 
				seu_states <= '1'; 
			when "010" => 				--b different
				current_state_corr <= current_state_a;
				seu_states <= '1'; 
			when "001" => 				--a different
				current_state_corr <= current_state_b;
				seu_states <= '1'; 
			when others => 			  -- errors in majority logic	
				current_state_corr <= IDLE;
				seu_states <= '1'; 
		end case;
	end process;   
	
	majority_logic_HDR_wait: process (HDR_wait_sig)

	begin

		case (HDR_wait_sig) is

			-- data = 0 - NO SEU
			when "000" => 	corr_data_HDR <= '0';  
							seu_HDR_wait <= '0';	
			
			-- data = 1 - NO SEU
			when "111" => 	corr_data_HDR <= '1';  
							seu_HDR_wait <= '0';
			
			-- data = 0 - SEU
			when "100" => 	corr_data_HDR <= '0';
							seu_HDR_wait <= '1';	
			when "010" =>	corr_data_HDR <= '0';
							seu_HDR_wait <= '1';
			when "001" => 	corr_data_HDR <= '0';
							seu_HDR_wait <= '1';
                  
			-- data = 1 - SEU            
			when "110" => 	corr_data_HDR <= '1';
							seu_HDR_wait <= '1';
			when "101" => 	corr_data_HDR <= '1';
							seu_HDR_wait <= '1';
			when "011" => 	corr_data_HDR <= '1'; 
							seu_HDR_wait <= '1';	
			
			when others => 	corr_data_HDR <= '0';
							seu_HDR_wait <= '0';
		end case;
			                                     
	end process;	
	
	
	seu <= seu_states or seu_HDR_wait; 
	HDR_wait <= corr_data_HDR;

end HDR_manager1;
