-- Synchronizer
-- 8x version				 
-- synchronizes the output clk on the THS channel in the input stream

LIBRARY IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_arith.all;
USE IEEE.std_logic_misc.all;
USE IEEE.std_logic_unsigned.all;
USE work.interfaces_pkg.all;

entity Synchronizer_8x_RHSEU is
	generic( 
		N_CYC : integer; -- = 4, 8, 16 respectively for tx_speed = 160, 320, 640 Mbit/s
		N_CYC_b : integer; -- = base-2 log of N_CYC
		CNT_mod : INTEGER;
		CNT_b : INTEGER;
		N1 : INTEGER;
		N2 : INTEGER);
	port(THS_det   : in  std_logic;	-- signals that a THS sequence has been detected
		clk         : in  std_logic;	-- input clk (clk_tx)
		rst_L  : in  std_logic; -- active low reset
		clk_out   : out  std_logic; -- output clk (clk40_RX)
		sync   : out  std_logic; -- signals that synchronization has been obtained
        resync : out  std_logic;
        seu : out STD_LOGIC

        );
end Synchronizer_8x_RHSEU;


architecture Synchronizer_8x_1 of Synchronizer_8x_RHSEU is
	
	component sync_counter_RHSEU
		generic(
			CNT_mod : INTEGER;
			CNT_b : INTEGER;
			N1 : INTEGER;
			N2 : INTEGER);
		port(
			CNT_en : in std_logic;
			clk : in std_logic;
			rst_CNT_L : in std_logic;
			rst_FF_L : in std_logic;
			rst_L : in std_logic;
			count_out : out std_logic_vector(CNT_b-1 downto 0);
			reset_sync : out std_logic;
			clr_all_CNT_L : out std_logic;
			sel : out std_logic;
			set_sync : out std_logic;
        seu : out STD_LOGIC
        );
	end component;
	
	component clk_splitter_8x_RHSEU 
		port(clk_in     : in  std_logic; 
			rst_L     : in  std_logic; 
			count : out std_logic_vector(2 downto 0); 
			clk_out     : out  std_logic_vector(7 downto 0);
        seu : out STD_LOGIC
            );
	end component;	
	
	COMPONENT dmux_8
	PORT
	(
		S		:	 IN STD_LOGIC_VECTOR(2 DOWNTO 0);
		I		:	 IN STD_LOGIC;
		O		:	 OUT STD_LOGIC_VECTOR(7 DOWNTO 0)
	);
	END COMPONENT; 

	component ffd_RHSEU
	port(
		CLR : in std_logic;
		en : in std_logic;
		clk : in std_logic;
		D : in std_logic;
		Q : out std_logic;
		QN : out std_logic;
        seu : out STD_LOGIC
        );
	end component; 

	
	signal clk_split : std_logic_vector(N_CYC-1 downto 0); -- output of clk splitter 
	signal clk_out_0 : std_logic_vector(N_CYC-1 downto 0); -- result of clk_split AND sel
	signal clk_out_int : std_logic; -- clk_out before DFF2
	signal clr_all_CNT_L : std_logic_vector(N_CYC-1 downto 0); -- clr_all_CNT_L outputs of sync_counters
	signal rst_all_CNT_L : std_logic; -- reset signal for all counters inside sync_counters
	signal rst_all_FF_L : std_logic; -- reset signal for all flip-flops inside sync_counters
	signal CNT_en : std_logic_vector(N_CYC-1 downto 0); -- enable input of sync_counters
	signal sel : std_logic_vector(N_CYC-1 downto 0); -- sel outputs of sync_counters
	signal sel_del : std_logic_vector(N_CYC-1 downto 0);
	signal set_sync : std_logic_vector(N_CYC-1 downto 0); -- set_sync outputs of sync_counters 
	signal reset_sync : std_logic_vector(N_CYC-1 downto 0); -- reset_sync outputs of sync_counters 
	signal demux_ctrl_in : std_logic_vector(N_CYC_b-1 downto 0);	-- count output of clk_splitter and control input of the demux
	signal sync_set : std_logic;	-- sets sync output
	signal sync_reset : std_logic; -- resets sync output 
	
	signal THS_det_del : std_logic; -- output of DFF													
	
	type stdlv_arr is array (N_CYC-1 downto 0) of std_logic_vector(CNT_b-1 downto 0);
	signal count : stdlv_arr;			-- for monitoring only
    
    signal seu_CLK_SPL, seu_DFF1, seu_DFF2 : STD_LOGIC;
	signal seu_CNT : std_logic_vector(N_CYC-1 downto 0);
    
begin
	
	CNT_GEN: for i in 0 to N_CYC-1 generate 
		
		CNT : sync_counter_RHSEU
		generic map (CNT_mod => CNT_mod, CNT_b => CNT_b, N1 => N1, N2 => N2)  
		port map
			(clk => clk,
			CNT_en => CNT_en(i),
			rst_CNT_L => rst_all_CNT_L,
			rst_FF_L => rst_all_FF_L,
			rst_L => rst_L,
			sel => sel(i),
			count_out => count(i),
			set_sync => set_sync(i),
			reset_sync => reset_sync(i),
			clr_all_CNT_L => clr_all_CNT_L(i),
            seu => seu_CNT(i)
            );
		
	end generate CNT_GEN;
	
	CLK_SPL : clk_splitter_8x_RHSEU
	port map
		(clk_in => clk,
		rst_L => rst_L,
		count => demux_ctrl_in,
		clk_out => clk_split,
       seu => seu_CLK_SPL
        );	
	
	-- demux providing the CNT_en input to sync_counters
	
	DMX:  dmux_8
	port map(
		O => CNT_en,
		S => demux_ctrl_in,
		I => THS_det_del
	);
	
	DFF1 : ffd_RHSEU
	port map(
		CLR => rst_L,
		en => '1',
		clk => clk,
		D => THS_det,
		Q => THS_det_del,
       seu => seu_DFF1
	);
	
	--------------------------------------
	
	RST_ALL_CNT: process(clr_all_CNT_L)	 -- generates rst_all_CNT_L as AND of all the bits of clr_all_CNT_L
		variable res : std_logic;
	begin
		res := '1';
		for i in 0 to N_CYC-1 loop
			res := res AND clr_all_CNT_L(i);
		end loop;
		rst_all_CNT_L <= res;
	end process;
	
	SYNCSET: process(set_sync)	 -- generates sync_set as OR of all the bits of set_sync
		variable res : std_logic;
	begin
		res := '0';
		for i in 0 to N_CYC-1 loop
			res := res OR set_sync(i);
		end loop;
		sync_set <= res;
	end process;
	
	rst_all_FF_L <= NOT sync_set;
	
	SYNCRESET: process(reset_sync)	 -- generates sync_reset as OR of all the bits of reset_sync
		variable res : std_logic;
	begin
		res := '0';
		for i in 0 to N_CYC-1 loop
			res := res OR reset_sync(i);
		end loop;
		sync_reset <= res;
	end process; 
	
	--------------------------------- clk_out generation
	clk_out_0 <= sel AND clk_split;
		
    clk_out_int <= or_reduce(clk_out_0);
    
    
	DFF2 : ffd_RHSEU
	port map(
		CLR => rst_L,
		en => '1',
		clk => clk,
		D => clk_out_int,
		Q => clk_out,
       seu => seu_DFF2
	);

	--------------------------------------
	
	SNC: process(clk, rst_L, set_sync, reset_sync) -- generates the sync output
		variable q: std_logic;
	begin  
		if rising_edge(clk) then
			if (rst_L AND NOT(sync_reset AND q)) = '0' then
				q := '0';
			elsif (sync_set AND NOT q) = '1' then
				q := '1';
			end if;
		end if;
			sync <= q;
	end process SNC;
	
	seu <= seu_CLK_SPL or seu_DFF1 or seu_DFF2 or or_reduce(seu_CNT);
 --------------------------------------
	-- generation of resync
    
	sel_del_proc : process(clk)
	begin  
      if rising_edge(clk) then
        if (rst_L = '0') then
          sel_del <= (OTHERS => '0') after 1 ps;
        else 
          sel_del <= sel after 1 ps;
        end if;
      end if;
	end process;
    
    resync <= '1' when not(sel = sel_del) else '0';
  
	   
end Synchronizer_8x_1; 
