-- THS_Detector
--					 
-- detects THS commands in the THS bits from the Deserializer

LIBRARY IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_arith.all;
USE IEEE.std_logic_unsigned.all;
use work.interfaces_pkg.all;

entity THS_Detector_RHSEU is
	port(THS_in   : in  std_logic_vector(N_THS-1 downto 0); -- THS bits to be checked
		clk         : in  std_logic;	-- input clock (must be connected to clk_tx)
		rst_L 		: in  std_logic;	-- active low reset
		sync   : in  std_logic; -- sync signal from Frame Synchronizer					   
		THS_valid   : in  std_logic; -- defines the window of validity of THS detection signals
		THS_det   : out  std_logic; -- signals that a THS sequence (without bit flips) has been detected
		TRG : out std_logic; -- signals that a TRG command (with a possible bit flip) has been detected
		HDR : out std_logic; -- signals that a HDR command (with a possible bit flip) has been detected
		THS_seu : out std_logic; -- signals that a bit flip has been detected in a THS sequence
       seu : out STD_LOGIC
       );
end THS_Detector_RHSEU;


architecture THS_Detector1 of THS_Detector_RHSEU is
	
	signal TRG_det : std_logic;	-- indicates that a TRG seq has been detected by the THS detector
	signal HDR_det : std_logic;	-- indicates that a HDR seq has been detected by the THS detector
	signal NOP_det : std_logic;	-- indicates that a NOP seq has been detected by the THS detector
	signal TRG_det_BF : std_logic;	-- indicates that a TRG seq with a bit flip has been detected by the THS detector_BF
	signal HDR_det_BF : std_logic;	-- indicates that a HDR seq with a bit flip has been detected by the THS detector_BF
	signal NOP_det_BF : std_logic;	-- indicates that a NOP seq with a bit flip has been detected by the THS detector_BF
	signal TRG_int : std_logic;	-- indicates that a TRG seq has been detected by the THS detector
	signal HDR_int : std_logic;	-- indicates that a HDR seq has been detected by the THS detector
	
	signal THS_enable : std_logic;	-- is pulled down after a THS detection to avoid other (false) detections for 3 clk40 cycles
	
	--signal NOP : std_logic;	-- indicates that a NOP seq (pure or with bit-flip) has been detected
	signal THS : std_logic;	-- indicates that a valid THS pattern has been detected
	
	component THS_enable_gen_RHSEU
		port(
			THS : in std_logic;
			THS_valid : in std_logic;
			clk : in std_logic;
			rst_L : in std_logic;
			THS_enable : out std_logic;
       seu : out STD_LOGIC
       );
	end component; 
	
	
begin
	
	-- THS_det combinatory logic: 
	
	TRG_det <= '1' when THS_in = "110001" 
	else '0';	  -- TRG, HDR and NOP sequences must be written here in reverse order (from bitN_THS-1 to bit0)
	
	HDR_det <= '1' when THS_in = "001101" 
	else '0';
	
	NOP_det <= '1' when THS_in = "101010" 
	else '0'; 
	
	THS_det <= TRG_det OR HDR_det OR NOP_det;
	----------------------------------------- 	 
	
	-- THS det_BF combinatory logic:
	
	TRG_det_BF <= '1' when 
	(THS_in = "010001") OR
	(THS_in = "100001") OR 
	(THS_in = "111001") OR
	(THS_in = "110101") OR 
	(THS_in = "110011") OR
	(THS_in = "110000") 
	else '0';
	
	HDR_det_BF <= '1' when
	(THS_in = "101101") OR 
	(THS_in = "011101") OR
	(THS_in = "000101") OR 
	(THS_in = "001001") OR
	(THS_in = "001111") OR 
	(THS_in = "001100") 
	else '0';
	
	NOP_det_BF <= '1' when
	(THS_in = "001010") OR 
	(THS_in = "111010") OR
	(THS_in = "100010") OR 
	(THS_in = "101110") OR
	(THS_in = "101000") OR 
	(THS_in = "101011") 
	else '0';
	
	-------------------------------------  
	
	TRG_int <= (TRG_det OR TRG_det_BF) AND THS_valid AND THS_enable;
	HDR_int <= (HDR_det OR HDR_det_BF) AND THS_valid AND THS_enable; --AND sync ;
	--NOP <= (NOP_det OR NOP_det_BF) AND sync AND THS_valid AND THS_enable;
	
	TRG <= TRG_int;
	HDR <= HDR_int;
	
	THS <= TRG_int OR HDR_int;
	
    THS_seu <= (TRG_det_BF or HDR_det_BF or NOP_det_BF) and THS_valid and THS_enable;
	----------------------------------------- 
	
	THS_EN_GEN : THS_enable_gen_RHSEU
	port map(
		THS => THS,
		THS_valid => THS_valid,
		clk => clk,
		rst_L => rst_L,
		THS_enable => THS_enable,
       seu => seu
		);
	
end THS_Detector1;
