-------------------------------------------------------------------------------
--
-- Title       : -- Title       : fb_control1_RHSEU
-- Design      : Interfaces
-- Author      : NC

-------------------------------------------------------------------------------
--
-- Description :  Triple Redundancy on States and Output Registers
--				  Protection against SEU in Flip-Flop
--				  Trg_manager is a registered Mealy FSM	(some registered outputs)
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use work.interfaces_pkg.all;  

entity FB_Control1_RHSEU is 
	port (
		clk: in STD_LOGIC;
		DV: in STD_LOGIC;
		Frm_Lim: in STD_LOGIC;
		GD: in STD_LOGIC;
		rst_L: in STD_LOGIC;
		LF: out STD_LOGIC;
		rst_WCNT_L: out STD_LOGIC;
		write_FLLF: out STD_LOGIC;
		write_LODT: out STD_LOGIC;
		seu: out STD_LOGIC);
end FB_Control1_RHSEU;

architecture FB_Control1_1 of FB_Control1_RHSEU is

type FB_CTRL1_Sreg0_type is (PKT_LOADING, IDLE, PKT_END);
signal next_state, current_state_a, current_state_b, current_state_c, current_state_corr: FB_CTRL1_Sreg0_type;
signal status :  STD_LOGIC_VECTOR(2 downto 0);
signal rst_WCNT_L_int : STD_LOGIC;
signal rst_WCNT_L_sig:  STD_LOGIC_VECTOR(0 to 2);
signal corr_data_rst_WCNT_L : STD_LOGIC;
signal a_eq_b: std_logic;
signal b_eq_c : std_logic; 
signal a_eq_c : std_logic;	
signal seu_states : std_logic;	
signal seu_rst_WCNT_L : std_logic;	

begin

fsm_state: process (clk)
begin
	if rising_edge(clk) then 
		if rst_L = '0' then	
			current_state_a <= IDLE;
			current_state_b <= IDLE;
			current_state_c <= IDLE;
		
			rst_WCNT_L_sig <= (others => '0');
		
		else	 
			current_state_a <= next_state;
			current_state_b <= next_state;
			current_state_c <= next_state;
			
            for i in 0 to 2 loop
                rst_WCNT_L_sig(i) <= rst_WCNT_L_int;
            end loop;
			
		end if;
	end if;
end process; 

fsm_comb_logic: process (current_state_corr, GD, Frm_Lim, DV)
begin
	case current_state_corr is
				when PKT_LOADING =>
					if DV = '0' then	
						next_state <= PKT_END;
						rst_WCNT_L_int <= '0';
					elsif (Frm_Lim = '1') AND (GD = '1') then	
						next_state <= PKT_LOADING;
						rst_WCNT_L_int <= '1';
					else
						next_state <= PKT_LOADING;
						rst_WCNT_L_int <= '1';
					end if;
				when IDLE =>
					if (DV = '1') AND (GD = '1') then		
						next_state <= PKT_LOADING;
						rst_WCNT_L_int <= '1';
                    else
						next_state <= IDLE;
						rst_WCNT_L_int <= '0';
					end if;
				when PKT_END =>
					if (DV = '1') AND (GD = '1') then	
						next_state <= PKT_LOADING;
						rst_WCNT_L_int <= '1';
					else
						next_state <= IDLE;
						rst_WCNT_L_int <= '0';
					end if;
				when others =>
						next_state <= IDLE;
						rst_WCNT_L_int <= '0';
			end case;
	end process;
	
  process (current_state_a, current_state_b, current_state_c)
	begin
        
		if (current_state_a = current_state_b) then
			a_eq_b <= '1';
		else
			a_eq_b <= '0';
		end if;
			
		if (current_state_a = current_state_c) then
			a_eq_c <= '1';
		else
			a_eq_c <= '0';
		end if;
				
		if (current_state_b = current_state_c) then
			b_eq_c <= '1';
		else
			b_eq_c <= '0';
		end if;
  end process;
  
  status <= a_eq_b & a_eq_c & b_eq_c;
  
  process (status, current_state_a, current_state_b)
    begin
		case (status) is 
			when "111" =>     			--All equal
				current_state_corr <= current_state_a; 
				seu_states <= '0'; 
			when "100" =>    			--c different
				current_state_corr <= current_state_a; 
				seu_states <= '1'; 
			when "010" => 				--b different
				current_state_corr <= current_state_a;
				seu_states <= '1'; 
			when "001" => 				--a different
				current_state_corr <= current_state_b;
				seu_states <= '1'; 
			when others => 			  -- errors in majority logic	
				current_state_corr <= IDLE;
				seu_states <= '1'; 
		end case;
	end process;   

	majority_logic_rst_WCNT_L: process (rst_WCNT_L_sig)
	begin
		case (rst_WCNT_L_sig) is

			-- data = 0 - NO SEU
			when "000" => 	corr_data_rst_WCNT_L <= '0';  
							seu_rst_WCNT_L <= '0';	
			
			-- data = 1 - NO SEU
			when "111" => 	corr_data_rst_WCNT_L <= '1';  
							seu_rst_WCNT_L <= '0';
			
			-- data = 0 - SEU
			when "100" => 	corr_data_rst_WCNT_L <= '0';
							seu_rst_WCNT_L <= '1';	
			when "010" =>	corr_data_rst_WCNT_L <= '0';
							seu_rst_WCNT_L <= '1';
			when "001" => 	corr_data_rst_WCNT_L <= '0';
							seu_rst_WCNT_L <= '1';
                  
			-- data = 1 - SEU            
			when "110" => 	corr_data_rst_WCNT_L <= '1';
							seu_rst_WCNT_L <= '1';
			when "101" => 	corr_data_rst_WCNT_L <= '1';
							seu_rst_WCNT_L <= '1';
			when "011" => 	corr_data_rst_WCNT_L <= '1'; 
							seu_rst_WCNT_L <= '1';	
			
			when others => 	corr_data_rst_WCNT_L <= '0';
							seu_rst_WCNT_L <= '0';
		end case;	                                     
	end process;
	
-- signal assignment statements for combinatorial outputs
write_FLLF_assignment:
write_FLLF <= '0' when (current_state_corr = IDLE and DV = '0') else
              '0' when (current_state_corr = IDLE and ((DV = '1') AND (GD = '1') and not (DV = '0'))) else
              '0' when (current_state_corr = IDLE) else
              '1' when (current_state_corr = PKT_LOADING and DV = '0') else
              '1' when (current_state_corr = PKT_LOADING and ((DV = '1') AND (Frm_Lim = '1') AND (GD = '1') and not (DV = '0'))) else
              '0' when (current_state_corr = PKT_LOADING and ((DV = '1') AND ((Frm_Lim = '0') OR (GD = '0')) and not (DV = '0') and not ((DV = '1') AND (Frm_Lim = '1') AND (GD = '1')))) else
              '0' when (current_state_corr = PKT_END and ((DV = '1') AND (GD = '1'))) else
              '0' when (current_state_corr = PKT_END and (DV = '0' and not ((DV = '1') AND (GD = '1')))) else
              '0';

write_LODT_assignment:
write_LODT <= '0' when (current_state_corr = IDLE and DV = '0') else
              '1' when (current_state_corr = IDLE and ((DV = '1') AND (GD = '1') and not (DV = '0'))) else
              '0' when (current_state_corr = IDLE) else
              '0' when (current_state_corr = PKT_LOADING and DV = '0') else
              '0' when (current_state_corr = PKT_LOADING and ((DV = '1') AND (Frm_Lim = '1') AND (GD = '1') and not (DV = '0'))) else
              '0' when (current_state_corr = PKT_LOADING and ((DV = '1') AND ((Frm_Lim = '0') OR (GD = '0')) and not (DV = '0') and not ((DV = '1') AND (Frm_Lim = '1') AND (GD = '1')))) else
              '1' when (current_state_corr = PKT_END and ((DV = '1') AND (GD = '1'))) else
              '0' when (current_state_corr = PKT_END and (DV = '0' and not ((DV = '1') AND (GD = '1')))) else
              '0';

LF_assignment:
LF <= '0' when (current_state_corr = IDLE and DV = '0') else
      '0' when (current_state_corr = IDLE and ((DV = '1') AND (GD = '1') and not (DV = '0'))) else
      '0' when (current_state_corr = IDLE) else
      '1' when (current_state_corr = PKT_LOADING and DV = '0') else
      '0' when (current_state_corr = PKT_LOADING and ((DV = '1') AND (Frm_Lim = '1') AND (GD = '1') and not (DV = '0'))) else
      '0' when (current_state_corr = PKT_LOADING and ((DV = '1') AND ((Frm_Lim = '0') OR (GD = '0')) and not (DV = '0') and not ((DV = '1') AND (Frm_Lim = '1') AND (GD = '1')))) else
      '0' when (current_state_corr = PKT_END and ((DV = '1') AND (GD = '1'))) else
      '0' when (current_state_corr = PKT_END and (DV = '0' and not ((DV = '1') AND (GD = '1')))) else
      '0';

rst_WCNT_L <= corr_data_rst_WCNT_L;
      	  
seu <= seu_states or seu_rst_WCNT_L;


end FB_Control1_1;
