-------------------------------------------------------------------------------
-- Description : enable generator for FF_RX
--
-- generates the ff_rx_enable signal that is high only for the clk_rx cycle
-- on which clk_ref_rx_rec has its rising edge
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

-------------------------------------------------------------------------------

entity ff_rx_enable_gen is
	generic (  -- speed_ind = 2 for 4x version, = 3 for 8x version, = 4 for 16x version
      speed_ind : integer range 2 to 4 := 3
	);
	port (
	    -- FF_LYNX fast clock (4x, 8x or 16x)
        clk_rx              : in std_logic;
        -- reference clock reconstructed by FF_RX
	    clk_ref_rx_rec      : in std_logic;
	    
        -- synchronous reset
	    reset               : in std_logic;
        
	    ff_rx_enable        : out std_logic
    );
end ff_rx_enable_gen;

---------------------------------------------------------------

architecture ff_rx_enable_gen1 of ff_rx_enable_gen is

    
	signal clk_ref_rx_rec_del : std_logic;
	signal pulse              : std_logic;
	signal cnt_out            : std_logic_vector (speed_ind-1 downto 0);  	

  begin
  
    -- single-pulse generator ------------------------
    spg_proc : process(clk_rx)
    begin
      if rising_edge(clk_rx) then
        if (reset = '1') then
          clk_ref_rx_rec_del <= '0' after 1 ps;
        else 
          clk_ref_rx_rec_del <= clk_ref_rx_rec after 1 ps;
        end if;
      end if;
    end process;
    
    pulse <= clk_ref_rx_rec and (not clk_ref_rx_rec_del);
    
    ----------------------------------------------------
 
    -- counter -----------------------------------------
    cnt_proc : process(clk_rx)
    begin
      if rising_edge(clk_rx) then
        if (reset = '1') or (pulse = '1') then
          cnt_out <= (OTHERS => '0') after 1 ps;
        else 
          cnt_out <= (cnt_out + 1) after 1 ps;
        end if;
      end if;
    end process;
    
    ----------------------------------------------------
 
    gen_4x : if (speed_ind = 2) generate
      ff_rx_enable <= '1' when (cnt_out = 2) else '0';
    end generate;
    
    gen_8x : if (speed_ind = 3) generate
      ff_rx_enable <= '1' when (cnt_out = 6) else '0';
    end generate;

    gen_16x : if (speed_ind = 4) generate
      ff_rx_enable <= '1' when (cnt_out = 14) else '0';
    end generate;
    
end ff_rx_enable_gen1;